import json
import os
import pathlib
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, TYPE_CHECKING
from unittest import mock

import globus_sdk
from pytest_lazyfixture import lazy_fixture
import pytest
import responses
from click.testing import CliRunner

from gcs_migration.commands.gcs_migration import main
from gcs_migration.models import CILogonConfiguration, MigrationPlan

if TYPE_CHECKING:
    from ..fixtures.gsi_authz_config import CalloutLine
    from ..fixtures.auth_documents import (
        IdentityDocumentDict,
        IdentityProviderDocumentDict,
    )


@pytest.fixture
def mock_map_identities_environment(
    mock_get_transfer_api: Any,
    mock_get_auth_api: Any,
    migration_plan_cmd_create: MigrationPlan,
    cilogon_callout: "CalloutLine",
    authorization_response_success: Tuple[Optional[str], str, int],
    tmp_path: pathlib.Path,
) -> Generator[MigrationPlan, None, None]:
    # Replace the callout configuration with a CILogon configuration
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        mock_migration_load.return_value = migration_plan_cmd_create
        assert migration_plan_cmd_create.gsi_authz_configuration is not None
        migration_plan_cmd_create.gsi_authz_configuration.callouts = [
            CILogonConfiguration(callout=cilogon_callout.config)
        ]
        migration_plan_cmd_create.root = tmp_path
        os.mkdir(tmp_path / "migration_plan")
        yield migration_plan_cmd_create


@pytest.fixture
def auth_resolve_osu_identity(mock_get_auth_api: Any) -> None:
    ac = globus_sdk.AuthClient()

    identity_provider_document: "IdentityProviderDocumentDict" = {
        "id": "309bf0b9-1549-43da-97c2-2bfdacacaa75",
        "name": "OSU EDU",
        "short_name": "Ohio State University",
        "alternative_names": [],
        "domains": ["osu.edu"],
    }

    identity_document: "IdentityDocumentDict" = {
        "username": "user@osu.edu",
        "status": "used",
        "name": "User Document",
        "id": "64620d3a-f6d6-4e56-9b48-a73b3ac5f070",
        "organization": "Ohio State University",
        "email": "user_email@osu.edu",
        "identity_type": "login",
        "identity_provider": identity_provider_document["id"],
    }

    # IdentityMap for migrating guest collections
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?usernames=user@osu.edu&provision=false",
        json={"identities": [identity_document]},
    )

    # Find usable domains in --domain values
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identity_providers?domains=osu.edu",
        json={"identity_providers": [identity_provider_document]},
    )


@pytest.fixture(params=["str", "file"])
def output_identity_mapping(
    tmp_path: pathlib.Path, request: Any
) -> Generator[Callable[[Dict[str, Any]], str], None, None]:
    def _output(identity_mapping: Dict[str, Any]) -> str:
        if request.param == "str":
            return json.dumps(identity_mapping)
        else:
            idmap_path = tmp_path / "idmap.json"
            with open(idmap_path, "w") as f:
                json.dump(identity_mapping, fp=f)
            return str(idmap_path)

    yield _output


@pytest.fixture
def good_identity_mapping(
    output_identity_mapping: Callable[[Dict[str, Any]], str]
) -> Generator[str, None, None]:
    yield output_identity_mapping(
        {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "mappings": [
                {
                    "source": "{username}",
                    "match": "(.*)@osu\\.edu",
                    "output": "{0}",
                }
            ],
        }
    )


@pytest.fixture
def bad_identity_mapping(
    output_identity_mapping: Callable[[Dict[str, Any]], str]
) -> Generator[str, None, None]:
    yield output_identity_mapping(
        {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "mappings": [
                {
                    "source": "{username}",
                    "match": "(.*)@osu\\.edu",
                }
            ],
        }
    )


@pytest.mark.parametrize(
    "organization", [None, "Ohio State University", "Unknown University"]
)
@pytest.mark.parametrize(
    "domains,local_user,expect",
    [
        (["osu.edu"], "user", 0),
        ([], "user", 0),
        (["osu.edu", "example.edu"], "user", 1),
    ],
)
def test_identity_mapping_cilogon(
    mock_map_identities_environment: MigrationPlan,
    cilogon_callout: "CalloutLine",
    auth_resolve_osu_identity: None,
    organization: Optional[str],
    domains: List[str],
    local_user: str,
    expect: int,
) -> None:
    migration_plan = mock_map_identities_environment

    if organization and "Unknown" in organization:
        expect = 1

    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    assert migration_plan.guest_collections is not None
    assert migration_plan.guest_collections[0] is not None
    assert migration_plan.guest_collections[0].credential is not None
    # Set mapped user to be the a user with an auth identity fixture
    migration_plan.guest_collections[0].credential.username = local_user

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "cilogon"]

    assert migration_plan.gsi_authz_configuration is not None
    assert migration_plan.gsi_authz_configuration.cilogon_callouts is not None
    assert migration_plan.gsi_authz_configuration.cilogon_callouts[0] is not None

    if expect == 0:
        assert migration_plan.guest_collections is not None
        assert len(migration_plan.guest_collections) > 0
        assert migration_plan.guest_collections[0].credential is not None
        migration_plan.guest_collections[0].credential.gsi_authz_callout_index = 0

    if organization:
        args.extend(["-o", organization])
    for d in domains:
        args.extend(["-d", d])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == expect

    if expect == 0:
        assert migration_plan.storage_gateway.allowed_domains is not None
        if domains:
            assert len(migration_plan.storage_gateway.allowed_domains) == len(domains)
            for d in domains:
                assert d in migration_plan.storage_gateway.allowed_domains
        else:
            assert "osu.edu" in migration_plan.storage_gateway.allowed_domains
        if "unknown" in local_user:
            assert "Unable to resolve" in result.output
    else:
        assert not migration_plan.storage_gateway.allowed_domains


@pytest.mark.parametrize(
    "organization", [None, "Ohio State University", "Unknown University"]
)
@pytest.mark.parametrize(
    "domains,local_user",
    [(["osu.edu"], "user"), ([], "user"), (["osu.edu", "example.edu"], "user")],
)
def test_identity_mapping_cilogon_no_config(
    mock_map_identities_environment: MigrationPlan,
    cilogon_callout: "CalloutLine",
    auth_resolve_osu_identity: None,
    organization: Optional[str],
    domains: List[str],
    local_user: str,
) -> None:
    migration_plan = mock_map_identities_environment

    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    assert migration_plan.guest_collections is not None
    assert migration_plan.guest_collections[0] is not None
    assert migration_plan.guest_collections[0].credential is not None
    # Set mapped user to be the a user with an auth identity fixture
    migration_plan.guest_collections[0].credential.username = local_user

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "cilogon"]

    assert migration_plan.gsi_authz_configuration is not None
    migration_plan.gsi_authz_configuration.callouts = []

    if organization:
        args.extend(["-o", organization])
    for d in domains:
        args.extend(["-d", d])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code != 0
    assert not migration_plan.storage_gateway.allowed_domains


@pytest.mark.parametrize(
    "domains",
    [["email.org"], ["unknown.org"]],
)
def test_unusable_domains(
    mock_map_identities_environment: MigrationPlan,
    domains: List[str],
) -> None:
    runner = CliRunner()
    args: List[str] = ["identity-mapping", "cilogon"]
    migration_plan = mock_map_identities_environment

    for d in domains:
        args.extend(["-d", d])

    assert migration_plan.storage_gateway is not None
    assert migration_plan.storage_gateway.allowed_domains is None

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == 1

    assert migration_plan.storage_gateway is not None
    assert migration_plan.storage_gateway.allowed_domains is None
    assert migration_plan.storage_gateway.identity_mappings is None


def test_requires_logged_in(mock_not_logged_in: None) -> None:
    runner = CliRunner()
    result = runner.invoke(
        main, ["identity-mapping", "cilogon"], catch_exceptions=False
    )
    assert "you will have to login" in result.output


@pytest.mark.parametrize(
    "mappings, expect",
    [
        (lazy_fixture("good_identity_mapping"), 0),
        (lazy_fixture("bad_identity_mapping"), 1),
    ],
)
def test_identity_mapping_cilogon_custom(
    mock_map_identities_environment: MigrationPlan,
    cilogon_callout: "CalloutLine",
    auth_resolve_osu_identity: None,
    mappings: str,
    expect: int,
) -> None:
    migration_plan = mock_map_identities_environment
    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    assert migration_plan.guest_collections is not None
    assert migration_plan.guest_collections[0] is not None
    assert migration_plan.guest_collections[0].credential is not None
    # Set mapped user to be the a user with an auth identity fixture
    migration_plan.guest_collections[0].credential.username = "user"

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "cilogon", "--identity-mapping", mappings]

    assert migration_plan.gsi_authz_configuration is not None
    assert migration_plan.gsi_authz_configuration.cilogon_callouts is not None
    assert migration_plan.gsi_authz_configuration.cilogon_callouts[0] is not None

    if expect == 0:
        assert migration_plan.guest_collections is not None
        assert len(migration_plan.guest_collections) > 0
        assert migration_plan.guest_collections[0].credential is not None
        migration_plan.guest_collections[0].credential.gsi_authz_callout_index = 0

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == expect

    if expect == 0:
        assert migration_plan.storage_gateway.allowed_domains is not None
        assert "osu.edu" in migration_plan.storage_gateway.allowed_domains
    else:
        assert not migration_plan.storage_gateway.allowed_domains
